/** 3DGPL *************************************************\
 * (MSDOS, i386+, VGA, WATCOM C)                          *
 * Header for hardware specific stuff.                    *
 *                                                        *
 * Defines:                                               *
 *  HW_init_screen           Opening output surface;      *
 *  HW_blit                  Colourmap onto the screen;   *
 *  HW_close_screen          Closing output;              *
 *                                                        *
 *  HW_init_event_loop       Runiing for events;          *
 *  HW_error                 Quiting with a message;      *
 *  HW_close_event_loop      Quiting running.             *
 *                                                        *
 * Internals:                                             *
 *  HWI_set_mode             Set hardware mode;           *
 *  HWI_set_colour           Set hardware colour.         *
 *                                                        *
 * (c) 1995-98 Sergei Savchenko, (savs@cs.mcgill.ca)      *
\**********************************************************/

#include "../hardware/hardware.h"           /* all the defenitions */
#include "../colour/colour.h"               /* CL_ stuff */
#include "../graphics/graphics.h"           /* G_ stuff */
#include <conio.h>                          /* kbhit() etc. */
#include <stdio.h>                          /* stderr */
#include <stdarg.h>                         /* var args */
#include <stdlib.h>                         /* exit */

int HW_running;                             /* event loop running? */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * Setting hardware screen mode 0x13.                    *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void HWI_set_mode(void);
#pragma aux HWI_set_mode=   \
 "mov  al,13h "            \
 "xor  ah,ah"              \
 "int  10h"                \
 modify [ax];

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * Setting hardware colour.                              *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void HWI_set_colour(int r,int g,int b,int number);
#pragma aux HWI_set_colour= \
 "mov  ax,1010h"            \
 "int  10h"                 \
 parm  [dh] [ch] [cl] [bx];

/**********************************************************\
 * Opening output and setting the palette.                *
\**********************************************************/

#define HW_TOTAL_COLOUR_CELLS 256           /* can't do more colours */

#if defined(_CI_)
void HW_init_screen(char *display_name,
                    char *screen_name
                   )
{
 int i;
 unsigned char r,g,b;

 HWI_set_mode();                            /* mode 13 */

 if(CL_colours->cl_no_colours>HW_TOTAL_COLOUR_CELLS)
  HW_error("(Hardware) Can't realize that big a palette.\n");

 for(i=0;i<CL_palette->cl_no_colours;i++)
 {
  r=CL_colours->cl_colours[i].cl_red>>2;
  g=CL_colours->cl_colours[i].cl_green>>2;
  b=CL_colours->cl_colours[i].cl_blue>>2;
  HWI_set_colour(r,g,b,i);                  /* palette */
 }
}
#endif
#if defined(_RGB_)
void HW_init_screen(char *display_name,
                    char *screen_name
                   )
{
 int i;
 unsigned char r,g,b;

 HWI_set_mode();                            /* mode 13 */

 for(i=0;i<HW_TOTAL_COLOUR_CELLS;i++)
 {
  r=CL_red(i)<<3;
  g=CL_green(i)<<4;
  b=CL_blue(i)<<3;

  HWI_set_colour(r,g,b,i);                  /* palette */
 }
}
#endif

/**********************************************************\
 * Colormap onto the screen.                              *
\**********************************************************/

void HW_blit(void)
{
 HW_copy_pixel(G_c_buffer,(HW_pixel*)0xa0000,G_c_buffer_size);
}

/**********************************************************\
 * Closing output.                                        *
\**********************************************************/

void HW_close_screen(void)
{
}

/**********************************************************\
 * Running the event loop.                                *
\**********************************************************/

void HW_init_event_loop(void (*application_main)(void),
                        void (*application_key_handler)(int key_code)
                       )
{
 char c;
 HW_running=1;

 while(HW_running==1)                       /* still running? hmmm */
 {
  if(kbhit())
  {
   c=getch();
   if(c==0) application_key_handler(getch());
   else application_key_handler(c);
  }
  application_main();
 }
}

/**********************************************************\
 * Stoping the event loop with a message.                 *
\**********************************************************/

void HW_error(char *s,...)
{
 va_list lst;

 va_start(lst,s);
 vfprintf(stderr,s,lst);                    /* printing a message */
 va_end(lst);
 HW_close_event_loop();                     /* quiting processing */
 exit(0);
}

/**********************************************************\
 * Stoping the event loop.                                *
\**********************************************************/

void HW_close_event_loop(void)
{
 HW_running=0;                              /* that's it */
}

/**********************************************************/
