/** 3DGPL *************************************************\
 * (WIN32, 386+, 16/32bit deep modes, 32bit compilers)    *
 * Header for hardware specific stuff.                    *
 *                                                        *
 * Ifdefs:                                                *
 *  _CI_                     Colour/Intensity model;      *
 *  _RGB_                    RGB model;                   *
 *  _8BPP_                   8 bit per pixel;             *
 *  _16BPP_                  16 bit per pixel;            *
 *  _32BPP_                  32 bit per pixel.            *
 *                                                        *
 * Defines:                                               *
 *  HW_set_int               Should be a macro;           *
 *  HW_set_16_bit            And this one.                *
 *                                                        *
 *  HW_init_screen           Opening output surface;      *
 *  HW_blit                  Colourmap onto the screen;   *
 *  HW_close_screen          Closing output;              *
 *                                                        *
 *  HW_init_event_loop       Runing for events;           *
 *  HW_error                 Quiting with a message;      *
 *  HW_close_event_loop      Quiting running.             *
 *                                                        *
 * Internals:                                             *
 *  HWI_null_function        An empty initial function.   *
 *                                                        *
 * (c) 1995-98 Sergei Savchenko, (savs@cs.mcgill.ca)      *
\**********************************************************/

#include "../hardware/hardware.h"           /* macros */
#include "../colour/colour.h"               /* CL_ stuff */
#include "../graphics/graphics.h"           /* G_ stuff */
#include <stdarg.h>                         /* var arg stuff */
#include <stdio.h>                          /* vsprintf */
#include <stdlib.h>                         /* exit */

extern int main(int n,char **o);            /* real main (well...) */

HDC HW_mem;                                 /* why? i wish i know why... */
HBITMAP HW_bmp;                             /* bitmap header */
RECT HW_rect;                               /* "client" area dimensions */
HWND HW_wnd;                                /* window */
HANDLE HW_instance;                         /* instance of this thing */
HPALETTE HW_palette;                        /* the palette headers */

int HW_cmd_show;                            /* not sure why, but needed */
char HW_class_name[]="Hardware";            /* what is class name anyway? */

struct                                      /* the logical palette */
{
 WORD hw_version;
 WORD hw_no_entries;
 PALETTEENTRY hw_entries[256];              /* how come it is 256 anyway? */
} HW_log_palette;

void HWI_null_function(void) {}             /* default handler */
void (*HW_application_main)(void)=HWI_null_function;
void (*HW_application_key_handler)(int key_code);

/**********************************************************\
 * Implementations for fast memory fill.                  *
\**********************************************************/

void HW_set_int(int *d,long l,int v)
{
 long i;

 for(i=0;i<l;i++) *d++=v;
}

/**********************************************************\
 * Implementations for another fast memory fill.          *
\**********************************************************/

void HW_set_16_bit(HW_16_bit *d,long l,HW_16_bit v)
{
 long i;

 for(i=0;i<l;i++) *d++=v;
}

/**********************************************************\
 * Creating a window.                                     *
\**********************************************************/

#if defined(_RGB_)
void HW_init_screen(char *display_name,
                    char *screen_name
                   )
#endif
{
 PAINTSTRUCT ps;
 int i,remap;

 HW_wnd=CreateWindow(HW_class_name,screen_name,WS_SYSMENU,
		     CW_USEDEFAULT,CW_USEDEFAULT,
		     HW_SCREEN_X_SIZE,
                     HW_SCREEN_Y_SIZE+GetSystemMetrics(SM_CYCAPTION),
		     NULL,NULL,HW_instance,NULL);

 HW_mem=CreateCompatibleDC(BeginPaint(HW_wnd,&ps));
 if(((GetDeviceCaps(ps.hdc,PLANES))!=1)||
    ((GetDeviceCaps(ps.hdc,BITSPIXEL))!=sizeof(HW_pixel)*8)
   )
  HW_error("(Hardware) I'd rather have %d bit screen.",
           sizeof(HW_pixel)*8
          );

 HW_bmp=CreateCompatibleBitmap(ps.hdc,HW_SCREEN_X_SIZE,HW_SCREEN_Y_SIZE);
 SelectObject(HW_mem,HW_bmp);

 EndPaint(HW_wnd,&ps);

 HW_rect.left=HW_rect.top=0;
 HW_rect.right=HW_SCREEN_X_SIZE;
 HW_rect.bottom=HW_SCREEN_Y_SIZE;

 ShowWindow(HW_wnd,HW_cmd_show);            /* generate messages */
 UpdateWindow(HW_wnd);
}

/**********************************************************\
 * Rendering the bitmap into the window.                  *
\**********************************************************/

void HW_blit(void)
{
 PAINTSTRUCT ps;

 BeginPaint(HW_wnd,&ps);

 SelectPalette(ps.hdc,HW_palette,FALSE);
 RealizePalette(ps.hdc);
 SetMapMode(ps.hdc,MM_TEXT);
 SetBitmapBits(HW_bmp,G_c_buffer_size*sizeof(HW_pixel),G_c_buffer);
 BitBlt(ps.hdc,0,0,HW_SCREEN_X_SIZE,HW_SCREEN_Y_SIZE,HW_mem,0,0,SRCCOPY);

 EndPaint(HW_wnd,&ps);
}

/**********************************************************\
 * Deallocating some stuff.                               *
\**********************************************************/

void HW_close_screen(void)
{
 DeleteDC(HW_mem);
 DeleteObject(HW_bmp);
}

/**********************************************************\
 * Running the event loop.                                *
\**********************************************************/

void HW_init_event_loop(void (*application_main)(void),
                        void (*application_key_handler)(int key_code)
                       )
{
 MSG msg;

 HW_application_main=application_main;
 HW_application_key_handler=application_key_handler;

 while(1)
 {
  if(PeekMessage(&msg,NULL,0,0,PM_REMOVE))  /* this IS sensitive part! */
  {
   if(msg.message == WM_QUIT) break;
   TranslateMessage(&msg);
   DispatchMessage(&msg);
  }
  else
  {
   InvalidateRect(HW_wnd,&HW_rect,TRUE);
   UpdateWindow(HW_wnd);
  }
 }
}

/**********************************************************\
 * Quiting with a message.                                *
\**********************************************************/

#define HW_MAX_ERROR_MESSAGE 256

void HW_error(char *s,...)
{
 char str[HW_MAX_ERROR_MESSAGE];
 va_list lst;

 va_start(lst,s);
 vsprintf(str,s,lst);                       /* forming the message */
 va_end(lst);

 MessageBox(NULL,str,"3Dgpl",MB_OK|MB_ICONSTOP|MB_SYSTEMMODAL);

 HW_close_event_loop();
 exit(0);                                   /* above might not be enough */
}

/**********************************************************\
 * Quitting the event loop.                               *
\**********************************************************/

void HW_close_event_loop(void)
{
 PostMessage(HW_wnd,WM_CLOSE,0,0L);         /* telling ourselves to quit */
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * Windows main callback function.                       *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * */

long FAR PASCAL WndProc(HWND hWnd,UINT message,
			WPARAM wParam,LPARAM lParam)
{
 switch(message)
 {
  case WM_PAINT:      HW_application_main();
                      break;
  case WM_ERASEBKGND: return(1L);          /* don't erase background */
  case WM_DESTROY:    PostQuitMessage(0);
                      break;
  case WM_KEYDOWN:    HW_application_key_handler(wParam);
                      break;
  default:            return(DefWindowProc(hWnd,message,wParam,lParam));
 }
 return(0L);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * Windows Main function.                                *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define HW_MAX_CLINE_OPTIONS 20

int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance,
		   LPSTR lpszCmdLine,int nCmdShow)
{
 WNDCLASS w;
 int n;
 char *start,*end;                          /* to get parametrs */
 char *o[HW_MAX_CLINE_OPTIONS];             /* to pass parameters */

 HW_cmd_show=nCmdShow;
 if((HW_instance=hPrevInstance)==NULL)
 {
  w.style=CS_HREDRAW|CS_VREDRAW;
  w.lpfnWndProc=(LPVOID)WndProc;
  w.cbClsExtra=0;
  w.cbWndExtra=0;
  w.hInstance=hInstance;
  w.hIcon=NULL;
  w.hCursor=NULL;
  w.hbrBackground=GetStockObject(WHITE_BRUSH);
  w.lpszMenuName=NULL;
  w.lpszClassName=HW_class_name;

  if(! RegisterClass(&w))
   return FALSE;
 }

 n=0;
 o[n++]="";
 start=lpszCmdLine;                         /* starting from here */
 while((end=strchr(start,' '))!=NULL)
 {
  if(n>=HW_MAX_CLINE_OPTIONS) 
   HW_error("(Hardware) Way too many command line options.\n");
  if(end!=start) o[n++]=start;              /* ignore empty ones */
  *end=0;                                   /* end of line */
  start=end+1;
 }
 if(strlen(start)>0) o[n++]=start;          /* the very last one */

 return(main(n,o));                         /* here, humane way to start */
}

/**********************************************************/
