/** 3DGPL *************************************************\
 * (X11, 8/16bit deep modes, 32bit mashine)               *
 * Header for hardware specific stuff.                    *
 *                                                        *
 * Ifdefs:                                                *
 *  _CI_                     Colour/Intensity model;      *
 *  _RGB_                    RGB model;                   *
 *  _8BPP_                   8 bit per pixel;             *
 *  _16BPP_                  16 bit per pixel;            *
 *  _32BPP_                  32 bit per pixel.            *
 *                                                        *
 * Defines:                                               *
 *  HW_set_int               Should be a macro;           *
 *  HW_set_16_bit            And this one.                *
 *                                                        *
 *  HW_init_screen           Opening output surface;      *
 *  HW_blit                  Colourmap onto the screen;   *
 *  HW_close_screen          Closing output;              *
 *                                                        *
 *  HW_init_event_loop       Runiing for events;          *
 *  HW_error                 Quiting with a message;      *
 *  HW_close_event_loop      Quiting running.             *
 *                                                        *
 * (c) 1995-98 Sergei Savchenko, (savs@cs.mcgill.ca)      *
\**********************************************************/

#include "../hardware/hardware.h"           /* harware dependent stuff */
#include "../colour/colour.h"               /* CL_ stuff */
#include "../graphics/graphics.h"           /* G_ stuff */
#include <X11/Xlib.h>                       /* most of X stuff */
#include <X11/Xutil.h>                      /* XSizeHints etc */
#include <stdarg.h>                         /* var args */
#include <stdio.h>                          /* stderr */
#include <stdlib.h>                         /* exit */

Display *HW_display;                        /* x server connection */
Visual *HW_visual;                          /* display's visual */
int HW_screen;                              /* deafualt screen */
Colormap HW_cmap;                           /* window's colourmap */
GC HW_gc;                                   /* default graphical conext */
Window HW_window;                           /* window being created */
Window HW_rootw;                            /* the very root window */
XColor HW_rgb[256];                         /* pixels rgb values */
unsigned long HW_cells[256];                /* colour numbers */
XImage *HW_c_buffer;                        /* the thing containing bitmap */
int HW_running;                             /* event loop still running */

/**********************************************************\
 * Implementations for fast memory fills.                 *
\**********************************************************/

void HW_set_int(int *d,long l,int v)
{
 long i;

 for(i=0;i<l;i++) *d++=v;
}

/**********************************************************\
 * Implementations for another fast memory fills.         *
\**********************************************************/

void HW_set_16_bit(HW_16_bit *d,long l,HW_16_bit v)
{
 long i;

 for(i=0;i<l;i++) *d++=v;
}

/**********************************************************\
 * Creating a window.                                     *
\**********************************************************/

#define HW_TOTAL_COLOUR_CELLS 256           /* can't set more colours */

void HW_init_screen(char *display_name,
                    char *screen_name
                   )
{
 XTextProperty name;
 XSizeHints *size;
 int i;

 if((HW_display=XOpenDisplay(display_name))==NULL)
  HW_error("(Hardware) Can't open specified display.\n");

 HW_screen=DefaultScreen(HW_display);
 HW_visual=DefaultVisual(HW_display,HW_screen);
 HW_rootw=RootWindow(HW_display,HW_screen);
 HW_gc=DefaultGC(HW_display,HW_screen);

 HW_window=XCreateSimpleWindow(HW_display,HW_rootw,0,0,
			       HW_SCREEN_X_SIZE,HW_SCREEN_Y_SIZE,
			       CopyFromParent,CopyFromParent,CopyFromParent
			      );

 XStringListToTextProperty((char**)&screen_name,1,&name);
 size=XAllocSizeHints();                    /* window manager hints */
 size->flags=PMinSize | PMaxSize;
 size->min_width=size->max_width=HW_SCREEN_X_SIZE;
 size->min_height=size->max_height=HW_SCREEN_Y_SIZE;
 XSetWMProperties(HW_display,HW_window,&name,&name,NULL,0,size,NULL,NULL);

 if(DefaultDepth(HW_display,HW_screen)!=sizeof(HW_pixel)*8)
  HW_error("(Hardware) I would rather have %d bit screen.\n",
           sizeof(HW_pixel)*8
          );

#if defined(_RGB_)
 #if defined(_8BPP_)
 HW_cmap=XCreateColormap(HW_display,HW_rootw,HW_visual,AllocNone);
 for(i=0;i<HW_TOTAL_COLOUR_CELLS;i++)
 {
  if(!XAllocColorCells(HW_display,HW_cmap,0,0,0,HW_cells+i,1))
   HW_error("(Hardware) Can't allocate a colour cell.\n");
  else
  {
   HW_rgb[i].flags=DoRed|DoGreen|DoBlue;
   HW_rgb[i].red=CL_red(i)<<13;
   HW_rgb[i].green=CL_green(i)<<14;
   HW_rgb[i].blue=CL_blue(i)<<13;
   HW_rgb[i].pixel=HW_cells[i];
  }
 }
 XStoreColors(HW_display,HW_cmap,HW_rgb,256);
 XSetWindowColormap(HW_display,HW_window,HW_cmap);

 #endif
#endif
#if defined(_CI_)
 #if defined(_8BPP_)

 HW_cmap=XCreateColormap(HW_display,HW_rootw,HW_visual,AllocNone);
 if(CL_colours->cl_no_colours>HW_TOTAL_COLOUR_CELLS)
  HW_error("(Hardware) Can't realize that big a palette.\n");

 for(i=0;i<CL_colours->cl_no_colours;i++)
 {
  if(!XAllocColorCells(HW_display,HW_cmap,0,0,0,HW_cells+i,1))
   HW_error("(Hardware) Can't allocate a colour cell.\n");
  else
  {
   HW_rgb[i].flags=DoRed|DoGreen|DoBlue;
   HW_rgb[i].red=((int)CL_colours->cl_colours[i].cl_red)<<8;
   HW_rgb[i].green=((int)CL_colours->cl_colours[i].cl_green)<<8;
   HW_rgb[i].blue=((int)CL_colours->cl_colours[i].cl_blue)<<8;
   HW_rgb[i].pixel=HW_cells[i];
  }
 }
 XStoreColors(HW_display,HW_cmap,HW_rgb,256);
 XSetWindowColormap(HW_display,HW_window,HW_cmap);
 #endif
#endif

 HW_c_buffer=XCreateImage(HW_display,HW_visual,sizeof(HW_pixel)*8,ZPixmap,
                          0,(char*)G_c_buffer,HW_SCREEN_X_SIZE,
                          HW_SCREEN_Y_SIZE,sizeof(HW_pixel)*8,
                          HW_SCREEN_X_SIZE*sizeof(HW_pixel)
                         );

 XMapWindow(HW_display,HW_window);
 XSync(HW_display,False);
 XSelectInput(HW_display,HW_window,KeyPressMask);
}

/**********************************************************\
 * Blitting a bitmap into the allocated window.           *
\**********************************************************/

void HW_blit(void)
{
 XPutImage(HW_display,HW_window,HW_gc,
           HW_c_buffer,0,0,0,0,HW_SCREEN_X_SIZE,HW_SCREEN_Y_SIZE
          );
}

/**********************************************************\
 * Deallocating a window.                                 *
\**********************************************************/

void HW_close_screen(void)
{
 XCloseDisplay(HW_display);
}

/**********************************************************\
 * The main event loop.                                   *
\**********************************************************/

void HW_init_event_loop(void (*application_main)(void),
                        void (*application_key_handler)(int key_code)
                       )
{
 XEvent report;                             /* beware putting XKeyEvent here */
 int key;

 HW_running=1;
 while(HW_running==1)
 {
  if(XCheckWindowEvent(HW_display,HW_window,KeyPressMask,&report)==1)
  {
   key=XKeycodeToKeysym(HW_display,((XKeyEvent*)&report)->keycode,0);
   if(key) application_key_handler(key);
  }
  application_main();
 }
}

/**********************************************************\
 * Quitng with a message.                                 *
\**********************************************************/

void HW_error(char *s,...)
{
 va_list lst;

 va_start(lst,s);
 vfprintf(stderr,s,lst);                    /* printing a message */
 va_end(lst);
 HW_close_event_loop();                     /* quiting processing */
 exit(0);
}

/**********************************************************\
 * Quiting the event loop.                                *
\**********************************************************/

void HW_close_event_loop(void)
{
 HW_running=0;
}

/**********************************************************/
